package cheddar.aadl;

import java.util.HashMap;

import org.osate.aadl2.instance.ComponentInstance;
import org.osate.aadl2.instance.ConnectionInstance;
import org.osate.aadl2.instance.ConnectionInstanceEnd;
import org.osate.aadl2.instance.FeatureInstance;
import org.osate.aadl2.instance.SystemInstance;
import org.osate.xtext.aadl2.properties.util.GetProperties;

import cheddar.adl.Address_Space;
import cheddar.adl.Aperiodic_Task;
import cheddar.adl.Core_Unit;
import cheddar.adl.Critical_Section;
import cheddar.adl.Dependency_Union;
import cheddar.adl.Generic_Resource;
import cheddar.adl.Generic_Task;
import cheddar.adl.Ipcp_Resource;
import cheddar.adl.Migrations_Type;
import cheddar.adl.Mono_Core_Processor;
import cheddar.adl.Np_Resource;
import cheddar.adl.Objects_Type;
import cheddar.adl.Pcp_Resource;
import cheddar.adl.Periodic_Task;
import cheddar.adl.Pip_Resource;
import cheddar.adl.Poisson_Task;
import cheddar.adl.Policies;
import cheddar.adl.Preemptives_Type;
import cheddar.adl.Resource_Dependency_Type;
import cheddar.adl.Resources_Type;
import cheddar.adl.Schedulers_Type;
import cheddar.adl.Scheduling_Parameters;
import cheddar.adl.Sporadic_Task;
import cheddar.adl.Tasks_Type;
import cheddar.osate2_plugin.Debug;


public class ToCheddarModel {

	// Cheddar entities to produce
	//
	public cheddar.adl.System result;

	// Cheddar unique object identifier
	//
	Cheddar_Id_Generator id;


	public ToCheddarModel() {
		id = new Cheddar_Id_Generator();
	}


	// Entry point to transform AADL => Cheddar ADL
	//
	public String parse(SystemInstance si) {

		result = new cheddar.adl.System();

		// Create Cheddar ADL instancies
		//
		for (ComponentInstance component : si.getAllComponentInstances()) {
			switch (component.getCategory()) {
			case PROCESSOR:
				Create_Uniprocessor_Component(result, component);
				break;
			case THREAD:
				Create_Thread_Component(result, component);
				break;
			case PROCESS:
				Create_Process_Component(result, component);
				break;
			case DATA:
				Create_Data_Component(result, component);
				break;
			default:
				break;
			}

		}

		// Create dependencies between Cheddar ADL instances
		//
		for (ComponentInstance component : si.getAllComponentInstances()) {
			switch (component.getCategory()) {
			case PROCESSOR:
				break;
			case THREAD:
				Thread_Component_Dependencies(result, component);
				break;
			case PROCESS:
				break;
			case DATA:
				Data_Component_Dependencies(result, component);
				break;
			default:
				break;
			}

		}


		// Produce Cheddar XML V3 and return it
		//
		return result.xmlString();
	}

	public void Thread_Component_Dependencies(cheddar.adl.System result, ComponentInstance component) {


		// Look for feature connections
		//
		for (FeatureInstance featureInstance : component.getFeatureInstances()) {
			for (ConnectionInstance connection : featureInstance.getDstConnectionInstances()) {
				 ConnectionInstanceEnd source = connection.getSource();
				 ConnectionInstanceEnd destination = connection.getDestination();

				// Add dependencies related to the resource
				//
				Generic_Task gt = result.Search_Task(component.getFullName().toString());
				Generic_Resource re = result.Search_Resource(source.getName());

				if ((gt != null) && (re != null))
				{
					Debug.Put("add resource dependency " + re.getName() + " " + gt.getName());

				Critical_Section cs = new Critical_Section();
				cs.setTaskBegin(1);
				cs.setTaskEnd(gt.getCapacity());
					HashMap<String, Critical_Section> css = re.getCriticalSections();
					if (css == null) {
						css = new HashMap<String, Critical_Section>();
						re.setCriticalSections(css);
					}
					css.put(gt.getName(), cs);

				Resource_Dependency_Type res_dep = new Resource_Dependency_Type();
				res_dep.setcheddarXmlId(id.getAndIncrement());
				res_dep.setResourceDependencyResource(re);
				res_dep.setResourceDependencyTask(gt);

				Dependency_Union dep = new Dependency_Union();
				dep.setResourceDependency(res_dep);

				result.dependencies.add(dep);
				}
			}
		}

	}

	public void Data_Component_Dependencies(cheddar.adl.System result, ComponentInstance component) {

	}

	public void Create_Uniprocessor_Component(cheddar.adl.System result, ComponentInstance component)
	{

		// Retrieve each property.
		// The policy followed here is to suppose
		// all properties are optionnal.
		// Then, each property is defined with a default value
		//
		String name = component.getFullName().toString();

		Integer quantum = Properties_With_Default_Value.getScheduler_Quantum(component);
		String source_text = Properties_With_Default_Value.getSource_Text(component);
		Schedulers_Type protocol = Properties_With_Default_Value.getSchedulingProtocol(component);
		Preemptives_Type preemptive = Preemptives_Type.Preemptive;
		if (!Properties_With_Default_Value.getPreemptive_Scheduler(component)) {
			preemptive = Preemptives_Type.Not_Preemptive;
		}

		// Instanciate Cheddar ADL entities
		//
		Scheduling_Parameters p = new Scheduling_Parameters();
		p.setPreemptiveType(preemptive);
		p.setSchedulerType(protocol);
		p.setUserDefinedSchedulerSourceFileName(source_text);
		p.setQuantum(quantum);

		Core_Unit c = new Core_Unit();
		c.setName("core_" + name);
		c.setScheduling(p);
		c.setcheddarXmlId(id.getAndIncrement());
		c.setObjectType(Objects_Type.Core_Object_Type);

		result.cores.add(c);

		// A processor, with its core unit
		Mono_Core_Processor mcp = new Mono_Core_Processor();
		mcp.setName(name);
		mcp.setcheddarXmlId(id.getAndIncrement());
		mcp.setObjectType(Objects_Type.Processor_Object_Type);
		mcp.setCore(c);
		mcp.setMigrationType(Migrations_Type.No_Migration_Type);

		Debug.Put("add core/processor " + mcp.getName());

		result.processors.add(mcp);

	}


	public void Create_Data_Component(cheddar.adl.System result, ComponentInstance component) {

		String name = component.getFullName().toString();

		Integer priority = Properties_With_Default_Value.getPriority(component);
		Resources_Type rt = Properties_With_Default_Value.getConcurrencyControlProtocol(component);

		String address_space_name = "";
		ComponentInstance co = component.getContainingComponentInstance();
		if (co != null) {
			address_space_name = co.getName();
		}
		co = GetProperties.getBoundProcessor(co);
		String cpu_name = "";
		if (co != null) {
			cpu_name = co.getName();
		}


		Generic_Resource gr;

		switch (rt) {
		case Immediate_Priority_Ceiling_Protocol:
			Ipcp_Resource ipcp = new Ipcp_Resource();
			gr = ipcp;
			break;
		case Priority_Inheritance_Protocol:
			Pip_Resource pip = new Pip_Resource();
			gr = pip;
			break;
		case Priority_Ceiling_Protocol:
			Pcp_Resource pcp = new Pcp_Resource();
			gr = pcp;
			break;
		// By default, Resources are No protocol resources
		default:
			Np_Resource np = new Np_Resource();
			gr = np;
			break;
		}

		gr.setcheddarXmlId(id.getAndIncrement());
		gr.setName(name);
		gr.setObjectType(Objects_Type.Resource_Object_Type);

		gr.setAddressSpaceName(address_space_name);
		gr.setCpuName(cpu_name);

		gr.setPriority(priority);
		gr.setProtocol(rt);
		gr.setState(1);

		Debug.Put("add resource " + gr.getName());

		result.resources.add(gr);

	}

	public void Create_Process_Component(cheddar.adl.System result, ComponentInstance component) {

		String name = component.getFullName().toString();

		Integer quantum = Properties_With_Default_Value.getScheduler_Quantum(component);
		String source_text = Properties_With_Default_Value.getSource_Text(component);
		Schedulers_Type protocol = Properties_With_Default_Value.getSchedulingProtocol(component);
		Preemptives_Type preemptive = Preemptives_Type.Preemptive;
		if (!Properties_With_Default_Value.getPreemptive_Scheduler(component)) {
			preemptive = Preemptives_Type.Not_Preemptive;
		}

		ComponentInstance co = GetProperties.getBoundProcessor(component);
		String cpu_name = "";
		if (co != null) {
			cpu_name = co.getName();
		}

		// Instanciate Cheddar ADL entities
		//
		Scheduling_Parameters p = new Scheduling_Parameters();
		p.setPreemptiveType(preemptive);
		p.setSchedulerType(protocol);
		p.setUserDefinedSchedulerSourceFileName(source_text);
		p.setQuantum(quantum);

		Address_Space c = new Address_Space();
		c.setName(name);
		c.setObjectType(Objects_Type.Address_Space_Object_Type);
		c.setScheduling(p);
		c.setcheddarXmlId(id.getAndIncrement());
		c.setCpuName(cpu_name);

		Debug.Put("add address space " + c.getName());

		result.address_spaces.add(c);

	}

	public void Create_Thread_Component(cheddar.adl.System result, ComponentInstance component) {

		String name = component.getFullName().toString();

		Integer priority = Properties_With_Default_Value.getPriority(component);
		Integer deadline = Properties_With_Default_Value.getDeadline(component);
		Integer period   =  Properties_With_Default_Value.getPeriod(component);
		Integer capacity = Properties_With_Default_Value.getExecutionTime(component);
		Integer jitter = Properties_With_Default_Value.getDispatch_Jitter(component);
		Integer start_time = Properties_With_Default_Value.getDispatch_Absolute_Time(component);
		Policies policy = Properties_With_Default_Value.getPOSIX_Scheduling_Policy(component);

		String address_space_name = "";
		ComponentInstance co = component.getContainingComponentInstance();
		if (co != null) {
			address_space_name = co.getName();
		}
		co = GetProperties.getBoundProcessor(co);
		String cpu_name = "";
		if (co != null) {
			cpu_name = co.getName();
		}

		Tasks_Type tt = Properties_With_Default_Value.getDispatchProtocol(component);

		Generic_Task gt;

		switch (tt) {
		case Aperiodic_Type:
			Aperiodic_Task at = new Aperiodic_Task();
			gt = at;
			break;
		case Sporadic_Type:
			Sporadic_Task st = new Sporadic_Task();
			st.setPeriod(period);
			st.setJitter(jitter);
			gt = st;
			break;
		case Poisson_Type:
			Poisson_Task pt = new Poisson_Task();
			pt.setPeriod(period);
			pt.setJitter(jitter);
			gt = pt;
			break;
		// By default, tasks are periodic
		default:
			Periodic_Task pt2 = new Periodic_Task();
			pt2.setPeriod(period);
			pt2.setJitter(jitter);
			gt = pt2;
			break;
        }

		gt.setcheddarXmlId(id.getAndIncrement());
		gt.setName(name);

		gt.setAddressSpaceName(address_space_name);
		gt.setCpuName(cpu_name);
		gt.setObjectType(Objects_Type.Task_Object_Type);

		gt.setCapacity(capacity);
		gt.setDeadline(deadline);
		gt.setPolicy(policy);
		gt.setPriority(priority);
		gt.setStartTime(start_time);

		Debug.Put("add task " + gt.getName());

		result.tasks.add(gt);
	}


}
